@@ -133,6 +133,26 @@ WriteArrayEntries(FILE *file, const statement_t *stats)
 	fprintf(file, "\t\t\t0,\n");
 }
 
+void
+WriteSubsystemServerRoutine(FILE *file, const char *typeModifier)
+{
+    fprintf(file, "%s mig_routine_t %s_routine\n", typeModifier,
+	    ServerDemux);
+    fprintf(file, "\t(const mach_msg_header_t *InHeadP)\n");
+
+    fprintf(file, "{\n");
+    fprintf(file, "\tint msgh_id;\n");
+    fprintf(file, "\n");
+    fprintf(file, "\tmsgh_id = InHeadP->msgh_id - %d;\n", SubsystemBase);
+    fprintf(file, "\n");
+    fprintf(file, "\tif ((msgh_id > %d) || (msgh_id < 0))\n",
+	    rtNumber - 1);
+    fprintf(file, "\t\treturn 0;\n");
+    fprintf(file, "\n");
+    fprintf(file, "\treturn %s_routines[msgh_id];\n", ServerDemux);
+    fprintf(file, "}\n");
+}
+
 static void
 WriteEpilog(FILE *file, const statement_t *stats)
 {
@@ -141,7 +161,7 @@ WriteEpilog(FILE *file, const statement_t *stats)
     /*
      * First, the symbol table
      */
-     fprintf(file, "static mig_routine_t %s_routines[] = {\n", ServerDemux);
+     fprintf(file, "mig_routine_t %s_routines[] = {\n", ServerDemux);
 
      WriteArrayEntries(file, stats);
 
@@ -196,25 +216,6 @@ WriteEpilog(FILE *file, const statement_t *stats)
     fprintf(file, "}\n");
     fprintf(file, "\n");
 
-    /*
-     * Then, the <subsystem>_server_routine routine
-     */
-    fprintf(file, "mig_external mig_routine_t %s_routine\n", ServerDemux);
-    fprintf(file, "\t(const mach_msg_header_t *InHeadP)\n");
-
-    fprintf(file, "{\n");
-    fprintf(file, "\tint msgh_id;\n");
-    fprintf(file, "\n");
-    fprintf(file, "\tmsgh_id = InHeadP->msgh_id - %d;\n", SubsystemBase);
-    fprintf(file, "\n");
-    fprintf(file, "\tif ((msgh_id > %d) || (msgh_id < 0))\n",
-	    rtNumber - 1);
-    fprintf(file, "\t\treturn 0;\n");
-    fprintf(file, "\n");
-    fprintf(file, "\treturn %s_routines[msgh_id];\n", ServerDemux);
-    fprintf(file, "}\n");
-    fprintf(file, "\n");
-
     /* symtab */
 
     if (GenSymTab) {
@@ -296,11 +297,11 @@ WriteVarDecls(FILE *file, const routine_t *rt)
     fprintf(file, "\n");
 
     if (!rt->rtSimpleFixedReply)
-	fprintf(file, "\tboolean_t msgh_simple;\n");
+	fprintf(file, "\tboolean_t msgh_simple = msgh_simple;\n");
     else if (!rt->rtSimpleCheckRequest)
     {
 	fprintf(file, "#if\tTypeCheck\n");
-	fprintf(file, "\tboolean_t msgh_simple;\n");
+	fprintf(file, "\tboolean_t msgh_simple = msgh_simple;\n");
 	fprintf(file, "#endif\t/* TypeCheck */\n");
 	fprintf(file, "\n");
     }
@@ -395,16 +396,16 @@ WriteCheckHead(FILE *file, const routine_t *rt)
 	fprintf(file, "\tmsgh_simple = !(In0P->Head.msgh_bits & MACH_MSGH_BITS_COMPLEX);\n");
 
     if (rt->rtNumRequestVar > 0)
-	fprintf(file, "\tif ((msgh_size < %d)",
+	fprintf(file, "\tif (mig_unlikely ((msgh_size < %d)",
 		rt->rtRequestSize);
     else
-	fprintf(file, "\tif ((In0P->Head.msgh_size != %d)",
+	fprintf(file, "\tif (mig_unlikely ((In0P->Head.msgh_size != %d)",
 		rt->rtRequestSize);
 
     if (rt->rtSimpleCheckRequest)
 	fprintf(file, " ||\n\t    %s(In0P->Head.msgh_bits & MACH_MSGH_BITS_COMPLEX)",
 		rt->rtSimpleReceiveRequest ? "" : "!");
-    fprintf(file, ")\n");
+    fprintf(file, "))\n");
     WriteMsgError(file, "MIG_BAD_ARGUMENTS");
     fprintf(file, "#endif\t/* TypeCheck */\n");
     fprintf(file, "\n");
@@ -422,7 +423,7 @@ WriteTypeCheck(FILE *file, const argument_t *arg)
 		arg->argRequestPos, arg->argTTName, arg->argVarName);
     else
     {
-	fprintf(file, "\tif (");
+	fprintf(file, "\tif (mig_unlikely (");
 	if (!it->itIndefinite) {
 	    fprintf(file, "(In%dP->%s%s.msgt_inline != %s) ||\n\t    ",
 		arg->argRequestPos, arg->argTTName,
@@ -450,7 +451,7 @@ WriteTypeCheck(FILE *file, const argument_t *arg)
 		    arg->argRequestPos, arg->argTTName,
 		    arg->argLongForm ? "l" : "",
 		    it->itNumber);
-	fprintf(file, "\t    (In%dP->%s.msgt%s_size != %d))\n",
+	fprintf(file, "\t    (In%dP->%s.msgt%s_size != %d)))\n",
 		arg->argRequestPos, arg->argTTName,
 		arg->argLongForm ? "l" : "",
 		it->itSize);
@@ -508,9 +509,10 @@ WriteCheckMsgSize(FILE *file, const argument_t *arg)
     if (arg->argRequestPos == rt->rtMaxRequestPos)
     {
 	fprintf(file, "#if\tTypeCheck\n");
-	fprintf(file, "\tif (msgh_size != %d + (", rt->rtRequestSize);
+	fprintf(file, "\tif (mig_unlikely (msgh_size != %d + (",
+		rt->rtRequestSize);
 	WriteCheckArgSize(file, arg);
-	fprintf(file, "))\n");
+	fprintf(file, ")))\n");
 
 	WriteMsgError(file, "MIG_BAD_ARGUMENTS");
 	fprintf(file, "#endif\t/* TypeCheck */\n");
@@ -538,10 +540,12 @@ WriteCheckMsgSize(FILE *file, const argument_t *arg)
 	   it won't underflow. */
 
 	if (LastVarArg)
-	    fprintf(file, "\tif (msgh_size != %d + msgh_size_delta)\n",
+	    fprintf(file,
+		    "\tif (mig_unlikely (msgh_size != %d + msgh_size_delta))\n",
 		rt->rtRequestSize);
 	else
-	    fprintf(file, "\tif (msgh_size < %d + msgh_size_delta)\n",
+	    fprintf(file,
+		    "\tif (mig_unlikely (msgh_size < %d + msgh_size_delta))\n",
 		rt->rtRequestSize);
 	WriteMsgError(file, "MIG_BAD_ARGUMENTS");
 
@@ -580,14 +584,45 @@ static void
 WriteExtractArgValue(FILE *file, const argument_t *arg)
 {
     const ipc_type_t *it = arg->argType;
+    boolean_t have_payload;
 
     if (arg->argMultiplier > 1)
 	WriteCopyType(file, it, "%s", "/* %s */ %s / %d",
 		      arg->argVarName, InArgMsgField(arg), arg->argMultiplier);
-    else if (it->itInTrans != strNULL)
-	WriteCopyType(file, it, "%s", "/* %s */ %s(%s)",
-		      arg->argVarName, it->itInTrans, InArgMsgField(arg));
-    else
+    else if ((have_payload = (it->itInTransPayload != strNULL &&
+	       strcmp(arg->argMsgField, "Head.msgh_request_port") == 0)) ||
+	     it->itInTrans != strNULL) {
+
+	if (have_payload) {
+	    argument_t argPayload = *arg;
+	    argPayload.argMsgField = "Head.msgh_bits";
+	    fprintf(file,
+		    "\tif (MACH_MSGH_BITS_LOCAL (%s) == "
+		    "MACH_MSG_TYPE_PROTECTED_PAYLOAD)\n"
+		    "\t", InArgMsgField(&argPayload));
+
+	    argPayload.argMsgField = "Head.msgh_protected_payload";
+	    WriteCopyType(file, it, "%s", "/* %s */ %s(%s)",
+			  arg->argVarName, it->itInTransPayload,
+			  InArgMsgField(&argPayload));
+
+	    fprintf(file,
+		    "\telse\n"
+		    "\t");
+
+	    if (it->itInTrans == strNULL)
+		fprintf(file, "\t%s = %s;",
+			arg->argVarName, InArgMsgField(arg));
+	    else
+		WriteCopyType(file, it, "%s", "/* %s */ %s(%s)",
+			      arg->argVarName, it->itInTrans,
+			      InArgMsgField(arg));
+	} else {
+	    WriteCopyType(file, it, "%s", "/* %s */ %s(%s)",
+			  arg->argVarName, it->itInTrans,
+			  InArgMsgField(arg));
+	}
+    } else
 	WriteCopyType(file, it, "%s", "/* %s */ %s",
 		      arg->argVarName, InArgMsgField(arg));
     fprintf(file, "\n");
@@ -736,9 +771,9 @@ WriteServerCallArg(FILE *file, const argument_t *arg)
 		    arg->argTTName,
 		    arg->argLongForm ? ".msgtl_header" : "");
 	    fprintf(file, "? %s ", InArgMsgField(arg));
-	    fprintf(file, ": *((%s **)%s)",
-			FetchServerType(arg->argType->itElement),
-			InArgMsgField(arg));
+	    fprintf(file, ": %s%s",
+		    InArgMsgField(arg),
+		    OOLPostfix);
 	}
 	else
 	    fprintf(file, "%s", InArgMsgField(arg));
@@ -775,8 +810,8 @@ WriteDestroyArg(FILE *file, const argument_t *arg)
 		arg->argRequestPos,
 		arg->argTTName,
 		arg->argLongForm ? ".msgtl_header" : "");
-	fprintf(file, "\t\t\t%smig_deallocate(* (vm_offset_t *) %s, ",
-		SubrPrefix, InArgMsgField(arg));
+	fprintf(file, "\t\t\t%smig_deallocate((vm_offset_t) %s%s, ",
+		SubrPrefix, InArgMsgField(arg), OOLPostfix);
 	if (multiplier > 1)
 	    fprintf(file, "%d * ", multiplier);
 	fprintf(file, " %s);\n", InArgMsgField(count));
@@ -905,6 +940,11 @@ WritePackArgValue(FILE *file, const argument_t *arg)
 		arg->argMsgField,
 		arg->argVarName,
 		it->itNumber);
+	    fprintf(file,
+		"\tif (OutP->%s < %d) OutP->%s += 1;\n",
+		arg->argCount->argMsgField,
+		it->itNumber,
+	        arg->argCount->argMsgField);
 	}
 	else {
 	    argument_t *count = arg->argCount;
@@ -932,9 +972,9 @@ WritePackArgValue(FILE *file, const argument_t *arg)
 			    arg->argTTName,
 			    arg->argLongForm ? ".msgtl_header" : "",
 			    arg->argDealloc->argVarName);
-		fprintf(file, "\t\t*((%s **)OutP->%s) = %sP;\n",
-			FetchServerType(btype),
+		fprintf(file, "\t\tOutP->%s%s = %sP;\n",
 			arg->argMsgField,
+			OOLPostfix,
 			arg->argVarName);
 		if (!arg->argRoutine->rtSimpleFixedReply)
 		    fprintf(file, "\t\tmsgh_simple = FALSE;\n");
@@ -1157,9 +1197,9 @@ WritePackArg(FILE *file, const argument_t *arg)
 			arg->argTTName,
 			arg->argLongForm ? ".msgtl_header" : "",
 			arg->argDealloc->argVarName);
-	    fprintf(file, "\t\t*((%s **)OutP->%s) = %sP;\n",
-			FetchServerType(it->itElement),
+	    fprintf(file, "\t\tOutP->%s%s = %sP;\n",
 			arg->argMsgField,
+			OOLPostfix,
 			arg->argVarName);
 	    if (!arg->argRoutine->rtSimpleFixedReply)
 		fprintf(file, "\t\tmsgh_simple = FALSE;\n");
@@ -1235,6 +1275,28 @@ WriteFieldDecl(FILE *file, const argument_t *arg)
     WriteFieldDeclPrim(file, arg, FetchServerType);
 }
 
+static void
+WriteDefaultRoutine(FILE *file, const routine_t *rt)
+{
+    fprintf(file, "\n/* Default implementation of %s */\n",
+	    rt->rtServerName);
+
+    fprintf(file, "#ifdef\tMIG_EOPNOTSUPP\n");
+
+    fprintf(file, "%s __attribute__ ((weak))\n%s\n",
+	    ReturnTypeStr(rt), rt->rtServerName);
+    fprintf(file, "(\n");
+    WriteList(file, rt->rtArgs, WriteServerVarDecl,
+	      akbServerArg, ",\n", "\n");
+
+    if (rt->rtReturn == argNULL)
+	fprintf(file, ") {}\n");
+    else
+	fprintf(file, ") { return MIG_EOPNOTSUPP; }\n");
+
+    fprintf(file, "#endif\t/* MIG_EOPNOTSUPP */\n");
+}
+
 static void
 WriteRoutine(FILE *file, const routine_t *rt)
 {
@@ -1303,6 +1365,8 @@ WriteRoutine(FILE *file, const routine_t *rt)
     }
 
     fprintf(file, "}\n");
+
+    WriteDefaultRoutine(file, rt);
 }
 
 void
